package djmil.cordacheckers.contracts;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleInputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.requireThat;

import java.util.List;

import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameResultState;
import net.corda.v5.base.annotations.ConstructorForDeserialization;
import net.corda.v5.base.annotations.CordaSerializable;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

public class GameBoardCommand implements Command {

    @CordaSerializable
    public static enum Type {
        MOVE,
        SURRENDER,
        FINISH;
    }

    private final Type type;
    private final List<Integer> move; // [0] from, [1] to

    // Serialisation service requires a default constructor
    public GameBoardCommand() {
        this.type = null;
        this.move = null;
    }

    @ConstructorForDeserialization
    public GameBoardCommand(Type type, List<Integer> move) {
        this.type = type;
        this.move = move;
    }

    public GameBoardCommand(Type type) {
        if (type == Type.MOVE)
            throw new CordaRuntimeException (BAD_ACTIONMOVE_CONSTRUCTOR);

        this.type = type;
        this.move = null;
    }

    public GameBoardCommand(List<Integer> move) {
        this.type = Type.MOVE;
        this.move = move;
    }

    public Type getType() {
        return this.type;
    }

    public List<Integer> getMove() {
        return this.move;
    }

    public MemberX500Name getInitiator(GameBoardState gameBoardState) {
        return gameBoardState.getMovePlayerName();
    }

    public MemberX500Name getRespondent(GameBoardState gameBoardState) {
        return gameBoardState.getCounterpartyName(gameBoardState.getMovePlayerName());
    }

    public static void validateSurrenderTrx(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, SURRENDER_INPUT_STATE);
        final var inGameBoardState = getSingleInputState(trx, GameBoardState.class);

        requireThat(trx.getOutputContractStates().size() == 1, SURRENDER_OUTPUT_STATE);
        final var outGameResultState = getSingleOutputState(trx, GameResultState.class);

        requireThat(inGameBoardState.getWhitePlayerName().compareTo(outGameResultState.getWhitePlayerName()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(inGameBoardState.getBlackPlayerName().compareTo(outGameResultState.getBlackPlayerName()) == 0, IN_OUT_PARTICIPANTS);
    }

    public static void validateMoveTrx(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, MOVE_INPUT_STATE);
        final var inGameBoardState = getSingleInputState(trx, GameBoardState.class);

        requireThat(trx.getOutputContractStates().size() == 1, MOVE_OUTPUT_STATE);
        final var outGameBoardState = getSingleOutputState(trx, GameBoardState.class);

        requireThat(inGameBoardState.getWhitePlayerName().compareTo(outGameBoardState.getWhitePlayerName()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(inGameBoardState.getBlackPlayerName().compareTo(outGameBoardState.getBlackPlayerName()) == 0, IN_OUT_PARTICIPANTS);
    }

    public static void validateFinishTrx(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, FINAL_MOVE_INPUT_STATE);
        final var inGameBoardState = getSingleInputState(trx, GameBoardState.class);

        requireThat(trx.getOutputContractStates().size() == 1, FINAL_MOVE_OUTPUT_STATE);
        final var outGameResultState = getSingleOutputState(trx, GameResultState.class);

        requireThat(inGameBoardState.getWhitePlayerName().compareTo(outGameResultState.getWhitePlayerName()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(inGameBoardState.getBlackPlayerName().compareTo(outGameResultState.getBlackPlayerName()) == 0, IN_OUT_PARTICIPANTS);
    }

    static final String BAD_ACTIONMOVE_CONSTRUCTOR = "Bad constructor for Action.MOVE";

    static final String SURRENDER_INPUT_STATE = "SURRENDER command should have exactly one GameBoardState input state";
    static final String SURRENDER_OUTPUT_STATE = "SURRENDER command should have exactly one GameResultState output state";

    static final String MOVE_INPUT_STATE = "MOVE command should have exactly one GameBoardState input state";
    static final String MOVE_OUTPUT_STATE = "MOVE command should have exactly one GameBoardState output state";
    static final String MOVE_OUT_BOARD = "MOVE command: move checkers rules violation";
    static final String MOVE_OUT_COLOR = "MOVE command: moveColor checkers rules violation";

    static final String FINAL_MOVE_INPUT_STATE = "FINAL_MOVE command should have exactly one GameBoardState input state";
    static final String FINAL_MOVE_OUTPUT_STATE = "FINAL_MOVE command should have exactly one GameResultState output state";

    static final String IN_OUT_PARTICIPANTS = "InputState and OutputState participants do not match";

    public static class CommandTypeException extends RuntimeException {
        public CommandTypeException() {
            super("Bad GameBoardCommand type");
        }
        public CommandTypeException(String message) {
            super(message);
        }
    }

}
